import unittest

from pytopojson import topology


class TopologyTestCase(unittest.TestCase):
    def setUp(self):
        self.topology = topology.Topology()

    def test_topology_exact_duplicate_lines_abc_abc_share_the_arc_abc(self):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[0, 0], [1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [0]},
                },
            },
            topology,
        )

    def test_topology_reversed_duplicate_lines_abc_cba_share_the_arc_abc(self):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[0, 0], [1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [~0]},
                },
            },
            topology,
        )

    def test_topology_when_old_arc_abc_extends_a_new_arc_ab_they_share_the_arc_ab(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [0]},
                },
            },
            topology,
        )

    def test_topology_when_a_reversed_old_arc_cba_extends_a_new_arc_ab_they_share_the_arc_ba(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[2, 0], [1, 0]], [[1, 0], [0, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [~1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_ade_shares_its_start_with_an_old_arc_abc_they_dont_share_arcs(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 1], [2, 1]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 1],
                "arcs": [[[0, 0], [1, 0], [2, 0]], [[0, 0], [1, 1], [2, 1]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_dec_shares_its_start_with_an_old_arc_abc_they_dont_share_arcs(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 1], [1, 1], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 1],
                "arcs": [[[0, 0], [1, 0], [2, 0]], [[0, 1], [1, 1], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abc_extends_an_old_arc_ab_they_share_the_arc_ab(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [0, 1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abc_extends_a_reversed_old_arc_ba_they_share_the_arc_ba(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[1, 0], [0, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[1, 0], [0, 0]], [[1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [~0, 1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_starts_bc_in_the_middle_of_an_old_arc_abc_they_share_the_arc_bc(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_bc_starts_in_the_middle_of_a_reversed_old_arc_cba_they_share_the_arc_cb(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
                "bar": {"type": "LineString", "coordinates": [[1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 0],
                "arcs": [[[2, 0], [1, 0]], [[1, 0], [0, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [~0]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abd_deviates_from_an_old_arc_abc_they_share_the_arc_ab(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 3, 0],
                "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]], [[1, 0], [3, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [0, 2]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abd_deviates_from_a_reversed_old_arc_cba_they_share_the_arc_ba(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [3, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 3, 0],
                "arcs": [[[2, 0], [1, 0]], [[1, 0], [0, 0]], [[1, 0], [3, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [~1, 2]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_dbc_merges_into_an_old_arc_abc_they_share_the_arc_bc(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[3, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 3, 0],
                "arcs": [[[0, 0], [1, 0]], [[1, 0], [2, 0]], [[3, 0], [1, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [2, 1]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_dbc_merges_into_a_reversed_old_arc_cba_they_share_the_arc_cb(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[2, 0], [1, 0], [0, 0]]},
                "bar": {"type": "LineString", "coordinates": [[3, 0], [1, 0], [2, 0]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 3, 0],
                "arcs": [[[2, 0], [1, 0]], [[1, 0], [0, 0]], [[3, 0], [1, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [2, ~0]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_dbe_shares_a_single_midpoint_with_an_old_arc_abc_they_share_the_point_b_but_no_arcs(
        self,
    ):
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 0], [2, 0]]},
                "bar": {"type": "LineString", "coordinates": [[0, 1], [1, 0], [2, 1]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 1],
                "arcs": [
                    [[0, 0], [1, 0]],
                    [[1, 0], [2, 0]],
                    [[0, 1], [1, 0]],
                    [[1, 0], [2, 1]],
                ],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1]},
                    "bar": {"type": "LineString", "arcs": [2, 3]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abde_skips_a_point_with_an_old_arc_abcde_they_share_arcs_ab_and_de(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]],
                },
                "bar": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [3, 0], [4, 0]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 4, 0],
                "arcs": [
                    [[0, 0], [1, 0]],
                    [[1, 0], [2, 0], [3, 0]],
                    [[3, 0], [4, 0]],
                    [[1, 0], [3, 0]],
                ],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1, 2]},
                    "bar": {"type": "LineString", "arcs": [0, 3, 2]},
                },
            },
            topology,
        )

    def test_topology_when_a_new_arc_abde_skips_a_point_with_a_reversed_old_arc_edcba_they_share_arcs_ba_and_ed(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[4, 0], [3, 0], [2, 0], [1, 0], [0, 0]],
                },
                "bar": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [3, 0], [4, 0]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 4, 0],
                "arcs": [
                    [[4, 0], [3, 0]],
                    [[3, 0], [2, 0], [1, 0]],
                    [[1, 0], [0, 0]],
                    [[1, 0], [3, 0]],
                ],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1, 2]},
                    "bar": {"type": "LineString", "arcs": [~2, 3, ~0]},
                },
            },
            topology,
        )

    def test_topology_when_an_arc_abcdbe_self_intersects_it_is_still_one_arc(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                }
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 4, 0],
                "arcs": [[[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]]],
                "objects": {"foo": {"type": "LineString", "arcs": [0]}},
            },
            topology,
        )

    def test_topology_when_an_old_arc_abcdbe_self_intersects_and_shares_a_point_b_the_old_arc_has_multiple_cuts(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                },
                "bar": {"type": "LineString", "coordinates": [[0, 1], [1, 0], [2, 1]]},
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 4, 1],
                "arcs": [
                    [[0, 0], [1, 0]],
                    [[1, 0], [2, 0], [3, 0], [1, 0]],
                    [[1, 0], [4, 0]],
                    [[0, 1], [1, 0]],
                    [[1, 0], [2, 1]],
                ],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0, 1, 2]},
                    "bar": {"type": "LineString", "arcs": [3, 4]},
                },
            },
            topology,
        )

    def test_topology_when_an_arc_abca_is_closed_it_has_one_arc(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]],
                }
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 1, 1],
                "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                "objects": {"foo": {"type": "LineString", "arcs": [0]}},
            },
            topology,
        )

    def test_topology_exact_duplicate_closed_lines_abca_and_abca_share_the_arc_abca(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]],
                },
                "bar": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 1, 1],
                "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [0]},
                },
            },
            topology,
        )

    def test_topology_reversed_duplicate_closed_lines_abca_and_acba_share_the_arc_abca(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [0, 1], [0, 0]],
                },
                "bar": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [0, 1], [1, 0], [0, 0]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 1, 1],
                "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                "objects": {
                    "foo": {"type": "LineString", "arcs": [0]},
                    "bar": {"type": "LineString", "arcs": [~0]},
                },
            },
            topology,
        )

    def test_topology_coincident_closed_polygons_abca_and_bcab_share_the_arc_bcab(self):
        topology = self.topology(
            {
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                },
                "bcab": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [0, 1], [0, 0], [1, 0]]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 1, 1],
                "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                "objects": {
                    "abca": {"type": "Polygon", "arcs": [[0]]},
                    "bcab": {"type": "Polygon", "arcs": [[0]]},
                },
            },
            topology,
        )

    def test_topology_coincident_reversed_closed_polygons_abca_and_bacb_share_the_arc_bcab(
        self,
    ):
        topology = self.topology(
            {
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                },
                "bacb": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [0, 0], [0, 1], [1, 0]]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 1, 1],
                "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                "objects": {
                    "abca": {"type": "Polygon", "arcs": [[0]]},
                    "bacb": {"type": "Polygon", "arcs": [[~0]]},
                },
            },
            topology,
        )

    def test_topology_coincident_closed_polygons_abca_and_dbed_share_the_point_b(self):
        topology = self.topology(
            {
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                },
                "dbed": {
                    "type": "Polygon",
                    "coordinates": [[[2, 1], [1, 0], [2, 2], [2, 1]]],
                },
            }
        )

        self.assertDictEqual(
            {
                "type": "Topology",
                "bbox": [0, 0, 2, 2],
                "arcs": [
                    [[1, 0], [0, 1], [0, 0], [1, 0]],
                    [[1, 0], [2, 2], [2, 1], [1, 0]],
                ],
                "objects": {
                    "abca": {"type": "Polygon", "arcs": [[0]]},
                    "dbed": {"type": "Polygon", "arcs": [[1]]},
                },
            },
            topology,
        )

    def test_topology_input_objects_are_mapped_to_topology_objects(self):
        """
        The topology `objects` is a map of geometry objects by name, allowing
        multiple GeoJSON geometry objects to share the same topology. When you
        pass multiple input files to bin/pytopojson, the basename of the file is
        used as the key, but you're welcome to edit the file to change it.
        """
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0.1, 0.2], [0.3, 0.4]]},
                "bar": {"type": "Polygon", "coordinates": [[[0.5, 0.6], [0.7, 0.8]]]},
            }
        )

        self.assertEqual(topology["objects"]["foo"]["type"], "LineString")
        self.assertEqual(topology["objects"]["bar"]["type"], "Polygon")

    def test_topology_features_are_mapped_to_geometries(self):
        """
        TopoJSON doesn't use features because you can represent the same
        information more compactly just by using geometry objects.
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                },
                "bar": {
                    "type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [[[0.5, 0.6], [0.7, 0.8]]],
                    },
                },
            }
        )

        self.assertEqual(topology["objects"]["foo"]["type"], "LineString")
        self.assertEqual(topology["objects"]["bar"]["type"], "Polygon")

    def test_topology_feature_collections_are_mapped_to_geometry_collections(self):
        topology = self.topology(
            {
                "collection": {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {
                                "type": "LineString",
                                "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                            },
                        },
                        {
                            "type": "Feature",
                            "geometry": {
                                "type": "Polygon",
                                "coordinates": [[[0.5, 0.6], [0.7, 0.8]]],
                            },
                        },
                    ],
                }
            }
        )

        self.assertEqual(
            topology["objects"]["collection"]["type"], "GeometryCollection"
        )
        self.assertEqual(len(topology["objects"]["collection"]["geometries"]), 2)
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][0]["type"], "LineString"
        )
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][1]["type"], "Polygon"
        )

    def test_topology_nested_geometry_collections(self):
        topology = self.topology(
            {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {
                            "type": "GeometryCollection",
                            "geometries": [
                                {
                                    "type": "LineString",
                                    "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                                }
                            ],
                        },
                        {"type": "Polygon", "coordinates": [[[0.5, 0.6], [0.7, 0.8]]]},
                    ],
                }
            }
        )

        self.assertEqual(
            len(
                topology["objects"]["collection"]["geometries"][0]["geometries"][0][
                    "arcs"
                ]
            ),
            1,
        )

    def test_topology_null_geometry_objects_are_preserved_in_geometry_collections(self):
        topology = self.topology(
            {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        None,
                        {"type": "Polygon", "coordinates": [[[0.5, 0.6], [0.7, 0.8]]]},
                    ],
                }
            }
        )

        self.assertEqual(
            topology["objects"]["collection"]["type"], "GeometryCollection"
        )
        self.assertEqual(len(topology["objects"]["collection"]["geometries"]), 2)
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][0]["type"], None
        )
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][1]["type"], "Polygon"
        )

    def test_topology_features_with_null_geometry_objects_are_preserved_in_feature_collections(
        self,
    ):
        topology = self.topology(
            {
                "collection": {
                    "type": "FeatureCollection",
                    "features": [
                        {"type": "Feature", "geometry": None},
                        {
                            "type": "Feature",
                            "geometry": {
                                "type": "Polygon",
                                "coordinates": [[[0.5, 0.6], [0.7, 0.8]]],
                            },
                        },
                    ],
                }
            }
        )

        self.assertEqual(
            topology["objects"]["collection"]["type"], "GeometryCollection"
        )
        self.assertEqual(len(topology["objects"]["collection"]["geometries"]), 2)
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][0]["type"], None
        )
        self.assertEqual(
            topology["objects"]["collection"]["geometries"][1]["type"], "Polygon"
        )

    def test_topology_top_level_features_with_null_geometry_objects_are_preserved(self):
        topology = self.topology({"feature": {"type": "Feature", "geometry": None}})
        self.assertDictEqual({"feature": {"type": None}}, topology["objects"])

    def test_topology_converting_a_feature_to_a_geometry_preserves_its_id(self):
        """
        To know what a geometry object represents, specify an id. I prefer
        numeric identifiers, such as ISO 3166-1 numeric, but strings work too.
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "id": 42,
                    "properties": {},
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )

        self.assertEqual(topology["objects"]["foo"]["type"], "LineString")
        self.assertEqual(topology["objects"]["foo"]["id"], 42)

    def test_topology_converting_a_feature_to_a_geometry_preserves_its_bbox(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "bbox": [0, 0, 10, 10],
                    "properties": {},
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )
        self.assertListEqual(topology["objects"]["foo"]["bbox"], [0, 0, 10, 10])

        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "properties": {},
                    "geometry": {
                        "type": "LineString",
                        "bbox": [0, 0, 10, 10],
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )
        self.assertListEqual(topology["objects"]["foo"]["bbox"], [0, 0, 10, 10])

    def test_topology_converting_a_feature_to_a_geometry_preserves_its_properties_but_only_if_non_empty(
        self,
    ):
        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "id": "Foo",
                    "properties": {"name": "George"},
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )
        self.assertDictEqual(
            topology["objects"]["foo"]["properties"], {"name": "George"}
        )

        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "id": "Foo",
                    "properties": {"name": "George"},
                    "geometry": {
                        "type": "GeometryCollection",
                        "geometries": [
                            {
                                "type": "LineString",
                                "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                            }
                        ],
                    },
                }
            }
        )
        self.assertDictEqual(
            topology["objects"]["foo"]["properties"], {"name": "George"}
        )

        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "id": "Foo",
                    "properties": {"name": "George", "demeanor": "curious"},
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )
        self.assertDictEqual(
            {"name": "George", "demeanor": "curious"},
            topology["objects"]["foo"]["properties"],
        )

        topology = self.topology(
            {
                "foo": {
                    "type": "Feature",
                    "id": "Foo",
                    "properties": {},
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[0.1, 0.2], [0.3, 0.4]],
                    },
                }
            }
        )
        self.assertEqual(None, topology["objects"]["foo"].get("properties", None))

    def test_topology_the_returned_transform_exactly_encompasses_the_input_geometry(
        self,
    ):
        """
        It's not required by the specification that the transform exactly
        encompass the input geometry, but this is a good test that the reference
        implementation is working correctly.
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [[1 / 8, 1 / 16], [1 / 2, 1 / 4]],
                }
            },
            quantization=2,
        )
        self.assertDictEqual(
            {"scale": [3 / 8, 3 / 16], "translate": [1 / 8, 1 / 16]},
            topology["transform"],
        )

        topology = self.topology(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [1 / 8, 1 / 16],
                            [1 / 2, 1 / 16],
                            [1 / 2, 1 / 4],
                            [1 / 8, 1 / 4],
                            [1 / 8, 1 / 16],
                        ]
                    ],
                }
            },
            quantization=2,
        )
        self.assertDictEqual(
            {"scale": [3 / 8, 3 / 16], "translate": [1 / 8, 1 / 16]},
            topology["transform"],
        )

    def test_topology_arc_coordinates_are_integers_with_delta_encoding(self):
        """
        TopoJSON uses integers with delta encoding to represent geometry
        efficiently. (Quantization is necessary for simplification anyway, so
        that we can identify which points are shared by contiguous geometry
        objects.) The delta encoding works particularly well because line strings
        are not random: most points are very close to their neighbors!
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [1 / 8, 1 / 16],
                            [1 / 2, 1 / 16],
                            [1 / 2, 1 / 4],
                            [1 / 8, 1 / 4],
                            [1 / 8, 1 / 16],
                        ]
                    ],
                }
            },
            quantization=2,
        )
        self.assertListEqual(
            [[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]], topology["arcs"][0]
        )

    def test_topology_points_coordinates_are_integers_with_delta_encoding(self):
        """
        TopoJSON uses integers with for points, also. However, there’s no delta-
        encoding, even for MultiPoints. And, unlike other geometry objects,
        points are still defined with coordinates rather than arcs.
        """
        topology = self.topology(
            {
                "foo": {"type": "Point", "coordinates": [1 / 8, 1 / 16]},
                "bar": {"type": "Point", "coordinates": [1 / 2, 1 / 4]},
            },
            quantization=2,
        )
        self.assertListEqual([], topology["arcs"])
        self.assertDictEqual(
            {"type": "Point", "coordinates": [0, 0]}, topology["objects"]["foo"]
        )
        self.assertDictEqual(
            {"type": "Point", "coordinates": [1, 1]}, topology["objects"]["bar"]
        )

        topology = self.topology(
            {
                "foo": {
                    "type": "MultiPoint",
                    "coordinates": [[1 / 8, 1 / 16], [1 / 2, 1 / 4]],
                }
            },
            quantization=2,
        )
        self.assertListEqual([], topology["arcs"])
        self.assertDictEqual(
            {"type": "MultiPoint", "coordinates": [[0, 0], [1, 1]]},
            topology["objects"]["foo"],
        )

    # def test_topology_quantization_rounds_to_the_closest_integer_coordinate_to_minimize_error(self):
    #     """
    #     Rounding is more accurate than flooring.
    #     """
    #     topology = self.topology({
    #         'foo': {
    #             'type': 'LineString',
    #             'coordinates': [
    #                 [['0'], ['0']],
    #                 [['5'], ['5']],
    #                 [['6'], ['6']],
    #                 [['0'], ['0']],
    #                 [['1'], ['1']],
    #                 [['9'], ['9']],
    #                 [['9'], ['9']],
    #                 [['5'], ['5']],
    #                 [['0'], ['0']],
    #                 [['4'], ['4']],
    #                 [['5'], ['5']],
    #                 [10, 10]
    #             ]
    #         }
    #     }, 11)
    #     self.assertListEqual([
    #         [0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9], [10, 10]
    #     ], client['feature'](topology, topology['objects']['foo']['geometry']['coordinates']))
    #     self.assertListEqual([
    #         [[0, 0], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1]]
    #     ], topology['arcs'])
    #     self.assertDictEqual({
    #         'scale': [1, 1],
    #         'translate': [0, 0]
    #     }, topology['transform'])

    # def test_topology_quantization_precisely_preserves_minimum_and_maximum_values(self):
    #     """
    #     When rounding, we must be careful not to exceed [±180°, ±90°]!
    #     """
    #     topology = self.topology({
    #         'foo': {
    #             'type': 'LineString',
    #             'coordinates': [[-180, -90], [0, 0], [180, 90]]
    #         }
    #     }, 3)
    #     self.assertListEqual([
    #         [-180, -90], [0, 0], [180, 90]
    #     ], client['feature'](topology, topology['objects']['foo']['geometry']['coordinates'])
    #     self.assertListEqual([
    #         [[0, 0], [1, 1], [1, 1]]
    #     ], topology['arcs'])
    #     self.assertDictEqual({
    #         'scale': [180, 90],
    #         'translate': [-180, -90]
    #     }, topology['transform'])

    def test_topology_precision_of_quantization_is_configurable(self):
        """
        GeoJSON inputs are in floating point format, so some error may creep in
        that prevents you from using exact match to determine shared points. The
        default quantization, 1e4, allows for 10,000 differentiable points in
        both dimensions. If you're using TopoJSON to represent especially high-
        precision geometry, you might want to increase the precision however,
        this necessarily increases the output size and the likelihood of seams
        between contiguous geometry after simplificatio[''] The quantization factor
        should be a power of ten for the most efficient representation, since
        JSON uses base-ten encoding for numbers.
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [
                        [1 / 8, 1 / 16],
                        [1 / 2, 1 / 16],
                        [1 / 8, 1 / 4],
                        [1 / 2, 1 / 4],
                    ],
                }
            },
            3,
        )
        self.assertListEqual([[0, 0], [2, 0], [-2, 2], [2, 0]], topology["arcs"][0])
        topology = self.topology(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [1 / 8, 1 / 16],
                            [1 / 2, 1 / 16],
                            [1 / 2, 1 / 4],
                            [1 / 8, 1 / 4],
                            [1 / 8, 1 / 16],
                        ]
                    ],
                }
            },
            5,
        )
        self.assertListEqual(
            [[0, 0], [4, 0], [0, 4], [-4, 0], [0, -4]], topology["arcs"][0]
        )

    def test_topology_coincident_points_are_removed(self):
        """
        Quantization may introduce coincident points, so these are remove['']
        """
        topology = self.topology(
            {
                "foo": {
                    "type": "LineString",
                    "coordinates": [
                        [1 / 8, 1 / 16],
                        [1 / 8, 1 / 16],
                        [1 / 2, 1 / 4],
                        [1 / 2, 1 / 4],
                    ],
                }
            },
            2,
        )
        self.assertListEqual([[[0, 0], [1, 1]]], topology["arcs"])
        topology = self.topology(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [1 / 8, 1 / 16],
                            [1 / 2, 1 / 16],
                            [1 / 2, 1 / 16],
                            [1 / 2, 1 / 4],
                            [1 / 8, 1 / 4],
                            [1 / 8, 1 / 4],
                            [1 / 8, 1 / 16],
                        ]
                    ],
                }
            },
            2,
        )
        self.assertListEqual(
            [[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]], topology["arcs"][0]
        )

    def test_topology_collapsed_lines_are_preserved(self):
        """
        Quantization may introduce degenerate features which have collapsed onto a single poin['']
        """
        topology = self.topology(
            {
                "foo": {"type": "LineString", "coordinates": [[0, 0], [1, 1], [2, 2]]},
                "bar": {
                    "type": "LineString",
                    "coordinates": [[-80, -80], [0, 0], [80, 80]],
                },
            },
            3,
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0]}, topology["objects"]["foo"]
        )
        self.assertListEqual([[1, 1], [0, 0]], topology["arcs"][0])

    def test_topology_collapsed_lines_in_a_multilinestring_are_preserved(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "MultiLineString",
                    "coordinates": [
                        [[1 / 8, 1 / 16], [1 / 2, 1 / 4]],
                        [[1 / 8, 1 / 16], [1 / 8, 1 / 16]],
                        [[1 / 2, 1 / 4], [1 / 8, 1 / 16]],
                    ],
                }
            },
            2,
        )
        self.assertEqual(len(topology["arcs"]), 2)
        self.assertListEqual([[0, 0], [0, 0]], topology["arcs"][1])
        self.assertListEqual([[0, 0], [1, 1]], topology["arcs"][0])
        self.assertListEqual([[0], [1], [~0]], topology["objects"]["foo"]["arcs"])

    def test_topology_collapsed_polygons_are_preserved(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]],
                },
                "bar": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [0, 1], [1, 1], [0, 0]]],
                },
                "baz": {
                    "type": "MultiPoint",
                    "coordinates": [[-80, -80], [0, 0], [80, 80]],
                },
            },
            3,
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0]]}, topology["objects"]["foo"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0]]}, topology["objects"]["bar"]
        )
        self.assertListEqual([[1, 1], [0, 0]], topology["arcs"][0])

    def test_topology_collapsed_polygons_in_a_MultiPolygon_are_preserved(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [
                                [1 / 8, 1 / 16],
                                [1 / 2, 1 / 16],
                                [1 / 2, 1 / 4],
                                [1 / 8, 1 / 4],
                                [1 / 8, 1 / 16],
                            ]
                        ],
                        [
                            [
                                [1 / 8, 1 / 16],
                                [1 / 8, 1 / 16],
                                [1 / 8, 1 / 16],
                                [1 / 8, 1 / 16],
                            ]
                        ],
                        [
                            [
                                [1 / 8, 1 / 16],
                                [1 / 8, 1 / 4],
                                [1 / 2, 1 / 4],
                                [1 / 2, 1 / 16],
                                [1 / 8, 1 / 16],
                            ]
                        ],
                    ],
                }
            },
            2,
        )
        self.assertTrue(0 < len(topology["arcs"]))
        self.assertTrue(2 <= len(topology["arcs"][0]))
        self.assertTrue(3 == len(topology["objects"]["foo"]["arcs"]))

    def test_topology_collapsed_geometries_in_a_GeometryCollection_are_preserved(self):
        topology = self.topology(
            {
                "collection": {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {"type": "MultiPolygon", "coordinates": []},
                        }
                    ],
                }
            },
            2,
        )
        self.assertTrue(len(topology["arcs"]) == 0)
        self.assertDictEqual(
            {
                "type": "GeometryCollection",
                "geometries": [{"type": "MultiPolygon", "arcs": []}],
            },
            topology["objects"]["collection"],
        )

    def test_topology_empty_geometries_are_not_removed(self):
        """
        If one of the top-level objects in the input is empty, however, it is
        still preserved in the output
        """
        topology = self.topology(
            {"foo": {"type": "MultiPolygon", "coordinates": []}}, 2
        )
        self.assertTrue(0 == len(topology["arcs"]))
        self.assertDictEqual(
            {"type": "MultiPolygon", "arcs": []}, topology["objects"]["foo"]
        )

    def test_topology_empty_polygons_are_not_removed(self):
        topology = self.topology(
            {
                "foo": {
                    "type": "FeatureCollection",
                    "features": [
                        {
                            "type": "Feature",
                            "geometry": {"type": "MultiPolygon", "coordinates": [[]]},
                        }
                    ],
                },
                "bar": {"type": "Polygon", "coordinates": []},
            }
        )
        self.assertTrue(0 == len(topology["arcs"]))
        self.assertDictEqual(
            {
                "type": "GeometryCollection",
                "geometries": [{"type": "MultiPolygon", "arcs": [[]]}],
            },
            topology["objects"]["foo"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": []}, topology["objects"]["bar"]
        )

    def test_topology_the_lines_ab_and_ab_share_the_same_arc(self):
        """

        A-----B

        """
        topology = self.topology(
            {
                "ab": {"type": "LineString", "coordinates": [[0, 0], [0, 1]]},
                "ba": {"type": "LineString", "coordinates": [[0, 0], [0, 1]]},
            }
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0]}, topology["objects"]["ab"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0]}, topology["objects"]["ba"]
        )

    def test_topology_the_lines_ab_and_ba_share_the_same_arc(self):
        """

        A-----B

        """
        topology = self.topology(
            {
                "ab": {"type": "LineString", "coordinates": [[0, 0], [0, 1]]},
                "ba": {"type": "LineString", "coordinates": [[0, 1], [0, 0]]},
            }
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0]}, topology["objects"]["ab"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [~0]}, topology["objects"]["ba"]
        )

    def test_topology_the_lines_acd_and_bcd_share_three_arcs(self):
        """

        A
         \
          \
           \
            \
             \
        B-----C-----D

        """
        topology = self.topology(
            {
                "acd": {"type": "LineString", "coordinates": [[0, 0], [1, 1], [2, 1]]},
                "bcd": {"type": "LineString", "coordinates": [[0, 1], [1, 1], [2, 1]]},
            }
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["acd"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [2, 1]}, topology["objects"]["bcd"]
        )

    def test_topology_the_lines_acd_and_dcb_share_three_arcs(self):
        """

        A
         \
          \
           \
            \
             \
        B-----C-----D

        """
        topology = self.topology(
            {
                "acd": {"type": "LineString", "coordinates": [[0, 0], [1, 1], [2, 1]]},
                "dcb": {"type": "LineString", "coordinates": [[2, 1], [1, 1], [0, 1]]},
            },
            quantization=3,
        )
        self.assertListEqual(
            [[[0, 0], [1, 2]], [[1, 2], [1, 0]], [[1, 2], [-1, 0]]],  # AC  # CD  # CB
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["acd"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [~1, 2]}, topology["objects"]["dcb"]
        )

    def test_topology_the_lines_acdf_and_bcdf_share_three_arcs(self):
        """

        A
         \
          \
           \
            \
             \
        B-----C-----D-----F
        """
        topology = self.topology(
            {
                "acdf": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 1]],
                },
                "bcdf": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 1]],
                },
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3]],  # AC
                [[1, 3], [1, 0], [1, 0]],  # CDF
                [[0, 3], [1, 0]],  # BC
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["acdf"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [2, 1]}, topology["objects"]["bcdf"]
        )

    def test_topology_the_lines_bcde_and_bcdf_share_three_arcs(self):
        """

                          E
                         /
                        /
                       /
                      /
                     /
        B-----C-----D-----F
        """
        topology = self.topology(
            {
                "bcde": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 0]],
                },
                "bcdf": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 1]],
                },
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 3], [1, 0], [1, 0]],  # BCD
                [[2, 3], [1, -3]],  # DE
                [[2, 3], [1, 0]],  # DF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["bcde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 2]}, topology["objects"]["bcdf"]
        )

    def test_topology_the_lines_acde_and_cd_share_three_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
              C-----D

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "cd": {"type": "LineString", "coordinates": [[1, 1], [2, 1]]},
            },
            4,
        )
        self.assertListEqual(
            [[[0, 0], [1, 3]], [[1, 3], [1, 0]], [[2, 3], [1, -3]]],  # AC  # CD  # DE
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1, 2]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [1]}, topology["objects"]["cd"]
        )

    def test_topology_the_lines_acde_and_bcd_share_four_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
        B-----C-----D

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "bcd": {"type": "LineString", "coordinates": [[0, 1], [1, 1], [2, 1]]},
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3]],  # AC
                [[1, 3], [1, 0]],  # CD
                [[2, 3], [1, -3]],  # DE
                [[0, 3], [1, 0]],  # BC
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1, 2]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [3, 1]}, topology["objects"]["bcd"]
        )

    def test_topology_the_lines_acde_and_cdf_share_four_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
              C-----D-----F

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "cdf": {"type": "LineString", "coordinates": [[1, 1], [2, 1], [3, 1]]},
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3]],  # AC
                [[1, 3], [1, 0]],  # CD
                [[2, 3], [1, -3]],  # DE
                [[2, 3], [1, 0]],  # CF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1, 2]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [1, 3]}, topology["objects"]["cdf"]
        )

    def test_topology_the_lines_acde_and_bcdf_share_five_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
        B-----C-----D-----F

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "bcdf": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 1]],
                },
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3]],  # AC
                [[1, 3], [1, 0]],  # CD
                [[2, 3], [1, -3]],  # DE
                [[0, 3], [1, 0]],  # BC
                [[2, 3], [1, 0]],  # DF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1, 2]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [3, 1, 4]}, topology["objects"]["bcdf"]
        )

    def test_topology_the_lines_acde_edca_and_acdf_share_three_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
              C-----D-----F

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "edca": {
                    "type": "LineString",
                    "coordinates": [[3, 0], [2, 1], [1, 1], [0, 0]],
                },
                "acdf": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 1]],
                },
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3], [1, 0]],  # ACD
                [[2, 3], [1, -3]],  # DE
                [[2, 3], [1, 0]],  # DF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 2]}, topology["objects"]["acdf"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [~1, ~0]}, topology["objects"]["edca"]
        )

    def test_topology_the_lines_acde_acdf_and_edca_share_three_arcs(self):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
              C-----D-----F

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "acdf": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 1]],
                },
                "edca": {
                    "type": "LineString",
                    "coordinates": [[3, 0], [2, 1], [1, 1], [0, 0]],
                },
            },
            4,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 3], [1, 0]],  # ACD
                [[2, 3], [1, -3]],  # DE
                [[2, 3], [1, 0]],  # DF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["acde"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 2]}, topology["objects"]["acdf"]
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [~1, ~0]}, topology["objects"]["edca"]
        )

    def test_topology_the_lines_acde_acdf_bcde_and_bcdf_and_their_reversals_share_five_arcs(
        self,
    ):
        r"""

        A                 E
         \               /
          \             /
           \           /
            \         /
             \       /
        B-----C-----D-----F

        """
        topology = self.topology(
            {
                "acde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 0]],
                },
                "acdf": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 1], [2, 1], [3, 1]],
                },
                "bcde": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 0]],
                },
                "bcdf": {
                    "type": "LineString",
                    "coordinates": [[0, 1], [1, 1], [2, 1], [3, 1]],
                },
                "edca": {
                    "type": "LineString",
                    "coordinates": [[3, 0], [2, 1], [1, 1], [0, 0]],
                },
                "fdca": {
                    "type": "LineString",
                    "coordinates": [[3, 1], [2, 1], [1, 1], [0, 0]],
                },
                "edcb": {
                    "type": "LineString",
                    "coordinates": [[3, 0], [2, 1], [1, 1], [0, 1]],
                },
                "fdcb": {
                    "type": "LineString",
                    "coordinates": [[3, 1], [2, 1], [1, 1], [0, 1]],
                },
            },
            4,
        )
        self.assertListEqual(
            topology["arcs"],
            [
                [[0, 0], [1, 3]],  # AC
                [[1, 3], [1, 0]],  # CD
                [[2, 3], [1, -3]],  # DE
                [[2, 3], [1, 0]],  # DF
                [[0, 3], [1, 0]],  # BC
            ],
        )
        self.assertDictEqual(
            topology["objects"]["acde"], {"type": "LineString", "arcs": [0, 1, 2]}
        )
        self.assertDictEqual(
            topology["objects"]["acdf"], {"type": "LineString", "arcs": [0, 1, 3]}
        )
        self.assertDictEqual(
            topology["objects"]["bcde"], {"type": "LineString", "arcs": [4, 1, 2]}
        )
        self.assertDictEqual(
            topology["objects"]["bcdf"], {"type": "LineString", "arcs": [4, 1, 3]}
        )
        self.assertDictEqual(
            topology["objects"]["edca"], {"type": "LineString", "arcs": [~2, ~1, ~0]}
        )
        self.assertDictEqual(
            topology["objects"]["fdca"], {"type": "LineString", "arcs": [~3, ~1, ~0]}
        )
        self.assertDictEqual(
            topology["objects"]["edcb"], {"type": "LineString", "arcs": [~2, ~1, ~4]}
        )
        self.assertDictEqual(
            topology["objects"]["fdcb"], {"type": "LineString", "arcs": [~3, ~1, ~4]}
        )

    def test_topology_the_polygons_abcda_and_befcb_share_three_arcs(self):
        """

        A-----B-----E
        |     |     |
        |     |     |
        |     |     |
        |     |     |
        |     |     |
        D-----C-----F

        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
                "befcb": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
                },
            },
            3,
        )
        self.assertListEqual(
            topology["arcs"],
            [
                [[1, 0], [0, 2]],  # BC
                [[1, 2], [-1, 0], [0, -2], [1, 0]],  # CDAB
                [[1, 0], [1, 0], [0, 2], [-1, 0]],  # BEFC
            ],
        )
        self.assertDictEqual(
            topology["objects"]["abcda"], {"type": "Polygon", "arcs": [[0, 1]]}
        )
        self.assertDictEqual(
            topology["objects"]["befcb"], {"type": "Polygon", "arcs": [[2, ~0]]}
        )

    def test_topology_the_polygons_abcda_and_abca_share_three_arcs(self):
        r"""

        A-----B
        |\    |
        | \   |
        |  \  |
        |   \ |
        |    \|
        D-----C

        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 0]]],
                },
            },
            2,
        )
        self.assertListEqual(
            [
                [[0, 0], [1, 0], [0, 1]],  # ABC
                [[1, 1], [-1, 0], [0, -1]],  # CDA
                [[1, 1], [-1, -1]],  # CA
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            topology["objects"]["abcda"], {"type": "Polygon", "arcs": [[0, 1]]}
        )
        self.assertDictEqual(
            topology["objects"]["abca"], {"type": "Polygon", "arcs": [[0, 2]]}
        )

    def test_topology_the_lines_abcde_and_abde_share_two_arcs(self):
        """

                    C
                   / \
                  /   \
                 /     \
                /       \
               /         \
        A-----B-----------D-----E

        """
        topology = self.topology(
            {
                "abcde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]],
                },
                "abde": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [3, 0], [4, 0]],
                },
            },
            5,
        )
        self.assertListEqual(
            topology["arcs"],
            [
                [[0, 0], [1, 0]],  # AB
                [[1, 0], [1, 0], [1, 0]],  # BCD
                [[3, 0], [1, 0]],  # DE
                [[1, 0], [2, 0]],  # BD
            ],
        )
        self.assertDictEqual(
            topology["objects"]["abcde"], {"type": "LineString", "arcs": [0, 1, 2]}
        )
        self.assertDictEqual(
            topology["objects"]["abde"], {"type": "LineString", "arcs": [0, 3, 2]}
        )

    def test_topology_the_polygons_abca_acda_and_bfcb_share_five_arcs(self):
        r"""

        A-----B
        |\    |\
        | \   | \
        |  \  |  \
        |   \ |   \
        |    \|    \
        D-----C-----F

        """
        topology = self.topology(
            {
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 0]]],
                },
                "acda": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 1], [0, 1], [0, 0]]],
                },
                "bfcb": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [2, 1], [1, 1], [1, 0]]],
                },
            },
            quantization=3,
        )

        self.assertListEqual(
            [
                [[0, 0], [1, 0]],  # AB
                [[1, 0], [0, 2]],  # BC
                [[1, 2], [-1, -2]],  # CA
                [[1, 2], [-1, 0], [0, -2]],  # CDA
                [[1, 0], [1, 2], [-1, 0]],  # BFC
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0, 1, 2]]}, topology["objects"]["abca"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[~2, 3]]}, topology["objects"]["acda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[4, ~1]]}, topology["objects"]["bfcb"]
        )

    def test_topology_the_polygons_abca_befcb_and_egfe_share_six_arcs(self):
        r"""

        A-----B-----E
         \    |     |\
          \   |     | \
           \  |     |  \
            \ |     |   \
             \|     |    \
              C-----F-----G

        """
        topology = self.topology(
            {
                "abca": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 0]]],
                },
                "befcb": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
                },
                "egfe": {
                    "type": "Polygon",
                    "coordinates": [[[2, 0], [3, 1], [2, 1], [2, 0]]],
                },
            },
            quantization=4,
        )

        self.assertListEqual(
            [
                [[1, 0], [0, 3]],  # BC
                [[1, 3], [-1, -3], [1, 0]],  # CAB
                [[1, 0], [1, 0]],  # BE
                [[2, 0], [0, 3]],  # EF
                [[2, 3], [-1, 0]],  # FC
                [[2, 0], [1, 3], [-1, 0]],  # EGF
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0, 1]]}, topology["objects"]["abca"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[2, 3, 4, ~0]]}, topology["objects"]["befcb"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[5, ~3]]}, topology["objects"]["egfe"]
        )

    def test_topology_the_polygons_abcda_abefgda_and_befgdcb_share_three_arcs(self):
        """
        //
        // A-----B-----E
        // |     |     |
        // |     |     |
        // D-----C     |
        // |           |
        // |           |
        // G-----------F
        //
        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
                "abefgda": {
                    "type": "Polygon",
                    "coordinates": [
                        [[0, 0], [1, 0], [2, 0], [2, 2], [0, 2], [0, 1], [0, 0]]
                    ],
                },
                "befgdcb": {
                    "type": "Polygon",
                    "coordinates": [
                        [[1, 0], [2, 0], [2, 2], [0, 2], [0, 1], [1, 1], [1, 0]]
                    ],
                },
            },
            quantization=3,
        )

        self.assertListEqual(
            [
                [[1, 0], [0, 1], [-1, 0]],  # BCD
                [[0, 1], [0, -1], [1, 0]],  # DAB
                [[1, 0], [1, 0], [0, 2], [-2, 0], [0, -1]],  # BEFGD
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0, 1]]}, topology["objects"]["abcda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[2, 1]]}, topology["objects"]["abefgda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[2, ~0]]}, topology["objects"]["befgdcb"]
        )

    def test_topology_the_polygons_abcda_and_bcdab_share_one_arc(self):
        """
        //
        // A-----B
        // |     |
        // |     |
        // D-----C
        //
        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
                "bcdab": {
                    "type": "Polygon",
                    "coordinates": [[[1, 0], [1, 1], [0, 1], [0, 0], [1, 0]]],
                },
            },
            quantization=2,
        )

        self.assertListEqual(
            [[[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]], topology["arcs"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0]]}, topology["objects"]["abcda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0]]}, topology["objects"]["bcdab"]
        )

    def test_topology_the_polygons_abcda_ehgfe_and_efghe_share_two_arcs(self):
        """
        //
        // A-----------------B
        // |                 |
        // |                 |
        // |     E-----F     |
        // |     |     |     |
        // |     |     |     |
        // |     H-----G     |
        // |                 |
        // |                 |
        // D-----------------C
        //
        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [
                        [[0, 0], [3, 0], [3, 3], [0, 3], [0, 0]],
                        [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                    ],
                },
                "efghe": {
                    "type": "Polygon",
                    "coordinates": [[[1, 1], [2, 1], [2, 2], [1, 2], [1, 1]]],
                },
            },
            quantization=4,
        )

        self.assertListEqual(
            [
                [[0, 0], [3, 0], [0, 3], [-3, 0], [0, -3]],  # ABCDA
                [[1, 1], [0, 1], [1, 0], [0, -1], [-1, 0]],  # EHGFE
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0], [1]]}, topology["objects"]["abcda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[~1]]}, topology["objects"]["efghe"]
        )

    def test_topology_the_polygons_abcda_ehgfe_and_fghef_share_two_arcs(self):
        """
        //
        // A-----------------B
        // |                 |
        // |                 |
        // |     E-----F     |
        // |     |     |     |
        // |     |     |     |
        // |     H-----G     |
        // |                 |
        // |                 |
        // D-----------------C
        //
        """
        topology = self.topology(
            {
                "abcda": {
                    "type": "Polygon",
                    "coordinates": [
                        [[0, 0], [3, 0], [3, 3], [0, 3], [0, 0]],
                        [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
                    ],
                },
                "fghef": {
                    "type": "Polygon",
                    "coordinates": [[[2, 1], [2, 2], [1, 2], [1, 1], [2, 1]]],
                },
            },
            quantization=4,
        )

        self.assertListEqual(
            [
                [[0, 0], [3, 0], [0, 3], [-3, 0], [0, -3]],  # ABCDA
                [[1, 1], [0, 1], [1, 0], [0, -1], [-1, 0]],  # EHGFE
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[0], [1]]}, topology["objects"]["abcda"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[~1]]}, topology["objects"]["fghef"]
        )

    def test_topology_the_polygon_bcdb_and_the_line_string_abe_share_three_arcs(self):
        r"""
           C-----D
            \   /
             \ /
        A-----B-----E
        """
        topology = self.topology(
            {
                "abe": {"type": "LineString", "coordinates": [[0, 1], [2, 1], [4, 1]]},
                "bcdb": {
                    "type": "Polygon",
                    "coordinates": [[[2, 1], [1, 0], [3, 0], [2, 1]]],
                },
            },
            quantization=5,
        )

        self.assertListEqual(
            [
                [[0, 4], [2, 0]],  # AB
                [[2, 4], [2, 0]],  # BE
                [[2, 4], [-1, -4], [2, 0], [-1, 4]],  # BCDB
            ],
            topology["arcs"],
        )
        self.assertDictEqual(
            {"type": "LineString", "arcs": [0, 1]}, topology["objects"]["abe"]
        )
        self.assertDictEqual(
            {"type": "Polygon", "arcs": [[2]]}, topology["objects"]["bcdb"]
        )



class Prequantize(object):
    def __init__(self):
        pass

    def __call__(self, objects, bbox, n, *args, **kwargs):
        self.quantize_geometry_type = {
            "GeometryCollection": self._geometry_collection_call,
            "Point": self._point_call,
            "MultiPoint": self._multi_point_call,
            "LineString": self._line_string_call,
            "MultiLineString": self._multi_line_string_call,
            "Polygon": self._polygon_call,
            "MultiPolygon": self._multi_polygon_call,
        }

        self.x_0, self.y_0, self.x_1, self.y_1 = bbox
        self.k_x = (n - 1) / (self.x_1 - self.x_0) if self.x_0 < self.x_1 else 1
        self.k_y = (n - 1) / (self.y_1 - self.y_0) if self.y_0 < self.y_1 else 1

        for k in objects:
            self.quantize_geometry(objects[k])

        return {
            "scale": [1 / self.k_x, 1 / self.k_y],
            "translate": [self.x_0, self.y_0],
        }

    def _geometry_collection_call(self, o):
        for geom in o["geometries"]:
            self.quantize_geometry(geom)

    def _point_call(self, o):
        o["coordinates"] = self.quantize_point(o["coordinates"])

    def _multi_point_call(self, o):
        o["coordinates"] = list(map(self.quantize_point, o["coordinates"]))

    def _line_string_call(self, o):
        o["arcs"] = self.quantize_line(o["arcs"])

    def _multi_line_string_call(self, o):
        o["arcs"] = list(map(self.quantize_line, o["arcs"]))

    def _polygon_call(self, o):
        o["arcs"] = self.quantize_polygon(o["arcs"])

    def _multi_polygon_call(self, o):
        o["arcs"] = list(map(self.quantize_polygon, o["arcs"]))

    def quantize_point(self, inp):
        return [
            round((inp[0] - self.x_0) * self.k_x),
            round((inp[1] - self.y_0) * self.k_y),
        ]

    def quantize_points(self, inp, m):
        i = j = 0
        n = len(inp)
        output = [None] * n
        p = [None, None]

        while i < n:
            p_i = inp[i]
            x = round((p_i[0] - self.x_0) * self.k_x)
            y = round((p_i[1] - self.y_0) * self.k_y)

            if p != [x, y]:
                p = [x, y]
                output[j] = p
                j += 1

            i += 1

        output = output[:j]

        while j < m:
            output.append([output[0][0], output[0][1]])
            j = len(output)

        return output

    def quantize_line(self, inp):
        return self.quantize_points(inp, 2)

    def quantize_ring(self, inp):
        return self.quantize_points(inp, 4)

    def quantize_polygon(self, inp):
        return list(map(self.quantize_ring, inp))

    def quantize_geometry(self, o):
        if o and o["type"] in self.quantize_geometry_type:
            self.quantize_geometry_type[o["type"]](o)



import unittest

from pytopojson.hash.hash import HashMap


def equal(a, b):
    return a is b


def hash(o):
    return o["hash"]


class HashMapTestCase(unittest.TestCase):
    def test_hashmap_can_get_an_object_by_key(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}
        map.set(key, 42)

        self.assertEqual(map.get(key), 42)

    def test_hashmap_get_returns_undefined_when_no_key_is_found(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.get(key), None)

    def test_hashmap_get_returns_the_missing_value_when_no_key_is_found(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.get(key, 42), 42)

    def test_hashmap_when_a_hash_collision_occurs_get_checks_that_the_keys_are_equal(
        self,
    ):
        map = HashMap(10, hash, equal)
        key_1 = {"hash": 1}
        key_2 = {"hash": 1}
        key_3 = {"hash": 1}

        map.set(key_1, "A")
        map.set(key_2, "B")

        self.assertEqual(map.get(key_1), "A")
        self.assertEqual(map.get(key_2), "B")
        self.assertEqual(map.get(key_3), None)

    def test_hashmap_can_set_an_object_by_key(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}

        map.set(key, 42)

        self.assertEqual(map.get(key), 42)

    def test_hashmap_can_set_an_object_by_key_if_not_already_set(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.maybe_set(key, 42), 42)
        self.assertEqual(map.get(key), 42)
        self.assertEqual(map.maybe_set(key, 43), 42)
        self.assertEqual(map.get(key), 42)

    def test_hashmap_set_returns_the_set_value(self):
        map = HashMap(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.set(key, 42), 42)

    def test_hashmap_set_throws_an_error_when_full(self):
        # minimum size of 16
        map = HashMap(0, hash, equal)
        keys = list()

        for i in range(16):
            keys.append({"hash": i})
            map.set(keys[i], True)

        # replacing is okay
        for i in range(16):
            map.set(keys[i], True)

        with self.assertRaises(ValueError):
            map.set({"hash": 16}, True)

    def test_hashmap_when_a_hash_collision_occurs_set_checks_that_the_keys_are_equal(
        self,
    ):
        map = HashMap(10, hash, equal)
        key_1 = {"hash": 1}
        key_2 = {"hash": 1}
        key_3 = {"hash": 1}

        self.assertEqual(map.set(key_1, "A"), "A")
        self.assertEqual(map.set(key_2, "B"), "B")
        self.assertEqual(map.get(key_1), "A")
        self.assertEqual(map.get(key_2), "B")
        self.assertEqual(map.get(key_1), "A")
        self.assertEqual(map.get(key_3), None)

    def test_hashmap_the_hash_function_must_return_a_nonnegative_integer_but_can_be_greater_than_size(
        self,
    ):
        map = HashMap(10, hash, equal)
        key = {"hash": 11}

        self.assertEqual(map.get(key), None)
        self.assertEqual(map.set(key, 42), 42)
        self.assertEqual(map.get(key), 42)



import numbers


__all__ = "in_delta"


def in_delta_dict(actual, expected, delta):
    for e in expected:
        if e not in actual or not in_delta(actual[e], expected[e], delta):
            return False
    for a in actual:
        if a not in expected:
            return False

    return True


def in_delta_array(actual, expected, delta):
    if len(actual) != len(expected):
        return False

    return all(
        map(
            lambda actual_i, expected_i: in_delta(actual_i, expected_i, delta),
            actual,
            expected,
        )
    )


def in_delta_number(actual, expected, delta):
    return (actual >= expected - delta) and (actual <= expected + delta)


def in_delta(actual, expected, delta=1e-6):
    if isinstance(expected, list):
        return isinstance(actual, list) and in_delta_array(actual, expected, delta)
    elif isinstance(expected, numbers.Number):
        return isinstance(actual, numbers.Number) and in_delta_number(
            actual, expected, delta
        )
    elif isinstance(expected, dict):
        return isinstance(actual, dict) and in_delta_dict(actual, expected, delta)
    else:
        return actual == expected



import unittest

from tests.hash.test_hashmap import equal, hash
from pytopojson.hash.hash import HashSet


def equal(a, b):
    return a is b


def hash(o):
    return o["hash"]


class HashSetTestCase(unittest.TestCase):
    def test_hashset_can_add_an_object(self):
        map = HashSet(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.add(key), True)
        self.assertEqual(map.has(key), True)

    def test_hashset_has_returns_false_when_no_key_is_found(self):
        map = HashSet(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.has(key), False)

    def test_hashset_when_a_hash_collision_occurs_get_checks_that_the_keys_are_equal(
        self,
    ):
        map = HashSet(10, hash, equal)
        key_1 = {"hash": 1}
        key_2 = {"hash": 1}
        key_3 = {"hash": 1}

        self.assertEqual(map.add(key_1), True)
        self.assertEqual(map.add(key_2), True)
        self.assertEqual(map.has(key_1), True)
        self.assertEqual(map.has(key_2), True)
        self.assertEqual(map.has(key_3), False)

    def test_hashset_add_returns_true(self):
        map = HashSet(10, hash, equal)
        key = {"hash": 1}

        self.assertEqual(map.add(key), True)

    def test_hashset_add_throws_an_error_when_full(self):
        # minimum size of 16
        map = HashSet(0, hash, equal)
        keys = list()

        for i in range(16):
            keys.append({"hash": i})
            map.add(keys[i])

        # replacing is okay
        for i in range(16):
            map.add(keys[i])

        with self.assertRaises(ValueError):
            map.add({"hash": 16})

    def test_hashset_the_hash_function_must_return_a_nonnegative_integer_but_can_be_greater_than_size(
        self,
    ):
        map = HashSet(10, hash, equal)
        key = {"hash": 11}

        self.assertEqual(map.has(key), False)
        self.assertEqual(map.add(key), True)
        self.assertEqual(map.has(key), True)


if __name__ == "__main__":
    unittest.main()



import json
import subprocess
import unittest

from tests import in_delta


class Geo2TopoTestCase(unittest.TestCase):
    def geo2topo(self, output, options=""):
        options = " ".join(options.split()).split(" ")
        p = subprocess.run(["python", "bin/geo2topo.py", *options], capture_output=True)
        out, err = p.stdout, p.stderr
        actual = json.loads(out.decode("utf8").strip().replace("'", '"'))
        with open(output, "r") as f:
            expected = json.load(f)

        self.assertTrue(in_delta.in_delta(actual, expected))

    def test_polygons(self):
        self.geo2topo(
            "tests/server/topojson/polygon-no-quantization.json",
            "polygon=tests/server/geojson/polygon-clockwise.json",
        )

    def test_quantized_polygons(self):
        self.geo2topo(
            "tests/server/topojson/polygon.json",
            "-q 1E4 polygon=tests/server/geojson/polygon-clockwise.json",
        )

    def test_empty_geometries(self):
        self.geo2topo(
            "tests/server/topojson/empty.json",
            "multilinestring=tests/server/geojson/empty-multilinestring.json"
            " multipoint=tests/server/geojson/empty-multipoint.json"
            " multipolygon=tests/server/geojson/empty-multipolygon.json"
            " multipolygon2=tests/server/geojson/empty-multipolygon2.json"
            " polygon=tests/server/geojson/empty-polygon.json",
        )

    def test_us_map(self):
        self.geo2topo(
            "tests/client/topojson/us-states.json",
            "tests/client/geojson/us-states.json",
        )

    def test_quantized_us_map(self):
        self.geo2topo(
            "tests/client/topojson/us-states-q1e2.json",
            "-q 1E2 tests/client/geojson/us-states.json",
        )



class Delta(object):
    def __init__(self):
        pass

    def __call__(self, arcs, *args, **kwargs):
        i = 0
        n = len(arcs)

        while i < n:
            arc = arcs[i]
            j = k = 1
            m = len(arc)
            point = arc[0]
            x_0, y_0 = point

            while j < m:
                point = arc[j]
                x_1, y_1 = point
                if x_0 != x_1 or y_0 != y_1:
                    arc[k] = [x_1 - x_0, y_1 - y_0]
                    x_0 = x_1
                    y_0 = y_1
                    k += 1

                j += 1

            # Each arc must be an array of two or more positions.
            if k == 1:
                arc[k] = [0, 0]
                k += 1

            arcs[i] = arc[:k]
            i += 1

        return arcs



import json
import math
import unittest

from pytopojson import (
    feature,
    mesh,
    stitch,
)


class MeshTestCase(unittest.TestCase):
    def setUp(self):
        self.feature = feature.Feature()
        self.mesh = mesh.Mesh()
        self.stitch = stitch.Stitch()

    def test_mesh_ignores_null_geometries(self):
        topology = {"type": "Topology", "objects": {}, "arcs": []}

        self.assertDictEqual(
            self.mesh(topology, [{"type": None}]),
            {"type": "MultiLineString", "coordinates": []},
        )

    def test_mesh_stitches_together_two_connected_line_strings(self):
        topology = {
            "type": "Topology",
            "objects": {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "arcs": [0]},
                        {"type": "LineString", "arcs": [1]},
                    ],
                }
            },
            "arcs": [[[1, 0], [2, 0]], [[0, 0], [1, 0]]],
        }
        self.assertDictEqual(
            {"type": "MultiLineString", "coordinates": [[[0, 0], [1, 0], [2, 0]]]},
            self.mesh(topology, topology["objects"]["collection"]),
        )

    def test_mesh_does_not_stitch_together_two_disconnected_line_strings(self):
        topology = {
            "type": "Topology",
            "objects": {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "arcs": [0]},
                        {"type": "LineString", "arcs": [1]},
                    ],
                }
            },
            "arcs": [[[2, 0], [3, 0]], [[0, 0], [1, 0]]],
        }
        self.assertDictEqual(
            {
                "type": "MultiLineString",
                "coordinates": [[[2, 0], [3, 0]], [[0, 0], [1, 0]]],
            },
            self.mesh(topology, topology["objects"]["collection"]),
        )



import unittest

from pytopojson import transform


class TransformTestCase(unittest.TestCase):
    def setUp(self):
        self.transform = transform.Transform()

    def test_transform_topology_returns_the_identity_function_if_transform_is_undefined(
        self,
    ):
        transform = self.transform(None)
        point = dict()
        self.assertDictEqual(transform(point), point)

    def test_transform_topology_returns_a_point_transform_function_if_transform_is_defined(
        self,
    ):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(transform([6, 7]), [16, 26])

    def test_transform_point_returns_a_new_point(self):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        point = [6, 7]
        self.assertListEqual(transform(point), [16, 26])
        self.assertListEqual(point, [6, 7])

    def test_transform_point_preserves_extra_dimensions(self):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(transform([6, 7, 42]), [16, 26, 42])
        self.assertListEqual(transform([6, 7, "foo"]), [16, 26, "foo"])
        self.assertListEqual(transform([6, 7, "foo", 42]), [16, 26, "foo", 42])

    def test_transform_point_transforms_individual_points(self):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(transform([1, 2]), [6, 11])
        self.assertListEqual(transform([3, 4]), [10, 17])
        self.assertListEqual(transform([5, 6]), [14, 23])

    def test_transform_point_index_transforms_delta_encoded_arcs(self):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(transform([1, 2], 0), [6, 11])
        self.assertListEqual(transform([3, 4], 1), [12, 23])
        self.assertListEqual(transform([5, 6], 2), [22, 41])
        self.assertListEqual(transform([1, 2], 3), [24, 47])
        self.assertListEqual(transform([3, 4], 4), [30, 59])
        self.assertListEqual(transform([5, 6], 5), [40, 77])

    def test_transform_point_index_transforms_multiple_delta_encoded_arcs(self):
        transform = self.transform({"scale": [2, 3], "translate": [4, 5]})
        self.assertListEqual(transform([1, 2], 0), [6, 11])
        self.assertListEqual(transform([3, 4], 1), [12, 23])
        self.assertListEqual(transform([5, 6], 2), [22, 41])
        self.assertListEqual(transform([1, 2], 0), [6, 11])
        self.assertListEqual(transform([3, 4], 1), [12, 23])
        self.assertListEqual(transform([5, 6], 2), [22, 41])



class Geometry(object):
    def __init__(self):
        pass

    def __call__(self, inputs, *args, **kwargs):
        self.output = dict()
        for k, v in inputs.items():
            self.output[k] = self.geomify_object(inputs[k])

        return self.output

    def geomify_object(self, input):
        if input is None:
            return {"type": None}
        elif input["type"] == "FeatureCollection":
            return self.geomify_feature_collection(input)
        elif input["type"] == "Feature":
            return self.geomify_feature(input)
        else:
            return self.geomify_geometry(input)

    def geomify_feature_collection(self, input):
        output = dict()
        output["type"] = "GeometryCollection"
        output["geometries"] = list(map(self.geomify_feature, input["features"]))
        if input.get("bbox", None) is not None:
            output["bbox"] = input["bbox"]

        return output

    def geomify_feature(self, input):
        output = self.geomify_geometry(input["geometry"])
        if input.get("id", None):
            output["id"] = input["id"]
        if input.get("bbox", None) is not None:
            output["bbox"] = input["bbox"]
        if input.get("properties", None):
            output["properties"] = input["properties"]

        return output

    def geomify_geometry(self, input):
        if input is None:
            return {"type": None}

        output = dict()
        if input["type"] == "GeometryCollection":
            output["type"] = "GeometryCollection"
            output["geometries"] = list(map(self.geomify_geometry, input["geometries"]))
        elif input["type"] == "Point" or input["type"] == "MultiPoint":
            output["type"] = input["type"]
            output["coordinates"] = input["coordinates"]
        else:
            output["type"] = input["type"]
            output["arcs"] = input["coordinates"]

        if input.get("bbox", None) is not None:
            output["bbox"] = input["bbox"]

        return output



from pytopojson import identity


class Untransform(object):
    def __init__(self):
        self.identity = identity.Identity()
        self.x_0 = 0
        self.y_0 = 0
        self.k_x = 0
        self.k_y = 0
        self.d_x = 0
        self.d_y = 0

    def __call__(self, transform=None, *args, **kwargs):
        if transform is None:
            return self.identity

        self.k_x, self.k_y = transform["scale"]
        self.d_x, self.d_y = transform["translate"]

        return self.func

    def func(self, input, i=None):
        if i is None or i == 0:
            self.x_0, self.y_0 = 0, 0
        output = input.copy()

        x_1 = int(round((input[0] - self.d_x) / self.k_x))
        y_1 = int(round((input[1] - self.d_y) / self.k_y))
        output[0] = x_1 - self.x_0
        output[1] = y_1 - self.y_0
        self.x_0 = x_1
        self.y_0 = y_1

        return output



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytopojson",
    version="1.1.1",
    author="Facundo Ferrín",
    author_email="facundo.ferrin@gmail.com",
    description="An extension to GeoJSON that encodes topology",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fferrin/pytopojson",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy >= 1.15.0",
    ],
    python_requires=">=3.7.3",
    entry_points={
        "console_scripts": [
            "geo2topo = bin.geo2topo:main",
            "topo2geo = bin.topo2geo:main",
            "topoquantize = bin.topoquantize:main",
        ]
    },
    scripts=[
        "bin/geo2topo.py",
        "bin/topo2geo.py",
        "bin/topoquantize.py",
    ],
)



import unittest

from pytopojson import (
    cut,
    extract,
)


class CutTestCase(unittest.TestCase):
    def setUp(self):
        self.cut = cut.Cut()
        self.extract = extract.Extract()

    def test_cut_exact_duplicate_lines_abc_and_abc_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abc2": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {"type": "LineString", "arcs": {0: 0, 1: 2}},
                "abc2": {"type": "LineString", "arcs": {0: 3, 1: 5}},
            },
            c["objects"],
        )

    def test_cut_reversed_duplicate_lines_abc_and_cba_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {"type": "LineString", "arcs": {0: 0, 1: 2}},
                "cba": {"type": "LineString", "arcs": {0: 3, 1: 5}},
            },
            c["objects"],
        )

    def test_cut_exact_duplicate_rings_abca_and_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "abca2": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [2, 0], [1, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "abca2": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_reversed_duplicate_rings_acba_and_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "acba": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [2, 0], [1, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "acba": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_rotated_duplicate_rings_bcab_and_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                    "bcab": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [2, 0], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "bcab": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_ring_abca_and_line_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abcaLine": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [0, 0]],
                    },
                    "abcaPolygon": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abcaLine": {"type": "LineString", "arcs": {0: 0, 1: 3}},
                "abcaPolygon": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_ring_bcab_and_line_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abcaLine": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [0, 0]],
                    },
                    "bcabPolygon": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [2, 0], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abcaLine": {"type": "LineString", "arcs": {0: 0, 1: 3}},
                "bcabPolygon": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_ring_abca_and_line_bcab_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "bcabLine": {
                        "type": "LineString",
                        "arcs": [[1, 0], [2, 0], [0, 0], [1, 0]],
                    },
                    "abcaPolygon": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [2, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "bcabLine": {"type": "LineString", "arcs": {0: 0, 1: 3}},
                "abcaPolygon": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_when_an_old_arc_abc_extends_a_new_arc_ab_abc_is_cut_into_ab_bc(self):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "ab": {"type": "LineString", "arcs": {0: 3, 1: 4}},
            },
            c["objects"],
        )

    def test_cut_when_a_reversed_old_arc_cba_extends_a_new_arc_ab_cba_is_cut_into_cb_ba(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "cba": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "ab": {"type": "LineString", "arcs": {0: 3, 1: 4}},
            },
            c["objects"],
        )

    def test_cut_when_a_new_arc_ade_shares_its_start_with_an_old_arc_abc_there_are_no_cuts(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "ade": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 1], [2, 1]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "ade": {"type": "LineString", "arcs": {0: 0, 1: 2}},
                "abc": {"type": "LineString", "arcs": {0: 3, 1: 5}},
            },
            c["objects"],
        )

    def test_cut_ring_aba_has_no_cuts(self):
        c = self.cut(
            self.extract(
                {"aba": {"type": "Polygon", "arcs": [[[0, 0], [1, 0], [0, 0]]]}}
            )
        )

        self.assertDictEqual(
            {"aba": {"type": "Polygon", "arcs": [{0: 0, 1: 2}]}}, c["objects"]
        )

    def test_cut_ring_aa_has_no_cuts(self):
        c = self.cut(
            self.extract({"aa": {"type": "Polygon", "arcs": [[[0, 0], [0, 0]]]}})
        )

        self.assertDictEqual(
            {"aa": {"type": "Polygon", "arcs": [{0: 0, 1: 1}]}}, c["objects"]
        )

    def test_cut_degenerate_ring_a_has_no_cuts(self):
        c = self.cut(self.extract({"a": {"type": "Polygon", "arcs": [[[0, 0]]]}}))

        self.assertDictEqual(
            {"a": {"type": "Polygon", "arcs": [{0: 0, 1: 0}]}}, c["objects"]
        )

    def test_cut_when_a_new_line_dec_shares_its_end_with_an_old_line_abc_there_are_no_cuts(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dec": {"type": "LineString", "arcs": [[0, 1], [1, 1], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {"type": "LineString", "arcs": {0: 0, 1: 2}},
                "dec": {"type": "LineString", "arcs": {0: 3, 1: 5}},
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abc_extends_an_old_line_ab_abc_is_cut_into_ab_bc(self):
        c = self.cut(
            self.extract(
                {
                    "ab": {"type": "LineString", "arcs": [[0, 0], [1, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "ab": {"type": "LineString", "arcs": {0: 0, 1: 1}},
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 2, 1: 3, "next": {0: 3, 1: 4}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abc_extends_a_reversed_old_line_ba_abc_is_cut_into_ab_bc(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "ba": {"type": "LineString", "arcs": [[1, 0], [0, 0]]},
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "ba": {"type": "LineString", "arcs": {0: 0, 1: 1}},
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 2, 1: 3, "next": {0: 3, 1: 4}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_starts_bc_in_the_middle_of_an_old_line_abc_abc_is_cut_into_ab_bc(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "bc": {"type": "LineString", "arcs": [[1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "bc": {"type": "LineString", "arcs": {0: 3, 1: 4}},
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_bc_starts_in_the_middle_of_a_reversed_old_line_cba_cba_is_cut_into_cb_ba(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "bc": {"type": "LineString", "arcs": [[1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "cba": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "bc": {"type": "LineString", "arcs": {0: 3, 1: 4}},
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abd_deviates_from_an_old_line_abc_abd_is_cut_into_ab_bd_and_abc_is_cut_into_ab_bc(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "abd": {"type": "LineString", "arcs": [[0, 0], [1, 0], [3, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "abd": {
                    "type": "LineString",
                    "arcs": {0: 3, 1: 4, "next": {0: 4, 1: 5}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abd_deviates_from_a_reversed_old_line_cba_cba_is_cut_into_cb_ba_and_abd_is_cut_into_ab_bd(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "abd": {"type": "LineString", "arcs": [[0, 0], [1, 0], [3, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "cba": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "abd": {
                    "type": "LineString",
                    "arcs": {0: 3, 1: 4, "next": {0: 4, 1: 5}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_dbc_merges_into_an_old_line_abc_dbc_is_cut_into_db_bc_and_abc_is_cut_into_ab_bc(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dbc": {"type": "LineString", "arcs": [[3, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "dbc": {
                    "type": "LineString",
                    "arcs": {0: 3, 1: 4, "next": {0: 4, 1: 5}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_dbc_merges_into_a_reversed_old_line_cba_dbc_is_cut_into_db_bc_and_cba_is_cut_into_cb_ba(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "cba": {"type": "LineString", "arcs": [[2, 0], [1, 0], [0, 0]]},
                    "dbc": {"type": "LineString", "arcs": [[3, 0], [1, 0], [2, 0]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "cba": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "dbc": {
                    "type": "LineString",
                    "arcs": {0: 3, 1: 4, "next": {0: 4, 1: 5}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_dbe_shares_a_single_midpoint_with_an_old_line_abc_dbe_is_cut_into_db_be_and_abc_is_cut_into_ab_bc(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abc": {"type": "LineString", "arcs": [[0, 0], [1, 0], [2, 0]]},
                    "dbe": {"type": "LineString", "arcs": [[0, 1], [1, 0], [2, 1]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abc": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 2}},
                },
                "dbe": {
                    "type": "LineString",
                    "arcs": {0: 3, 1: 4, "next": {0: 4, 1: 5}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abde_skips_a_point_with_an_old_line_abcde_abde_is_cut_into_ab_bd_de_and_abcde_is_cut_into_ab_bcd_de(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abcde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]],
                    },
                    "abde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [3, 0], [4, 0]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abcde": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 3, "next": {0: 3, 1: 4}}},
                },
                "abde": {
                    "type": "LineString",
                    "arcs": {0: 5, 1: 6, "next": {0: 6, 1: 7, "next": {0: 7, 1: 8}}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_new_line_abde_skips_a_point_with_a_reversed_old_line_edcba_abde_is_cut_into_ab_bd_de_and_edcba_is_cut_into_ed_dcb_ba(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "edcba": {
                        "type": "LineString",
                        "arcs": [[4, 0], [3, 0], [2, 0], [1, 0], [0, 0]],
                    },
                    "abde": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [3, 0], [4, 0]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "edcba": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 3, "next": {0: 3, 1: 4}}},
                },
                "abde": {
                    "type": "LineString",
                    "arcs": {0: 5, 1: 6, "next": {0: 6, 1: 7, "next": {0: 7, 1: 8}}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_line_abcdbe_self_intersects_with_its_middle_it_is_not_cut(self):
        c = self.cut(
            self.extract(
                {
                    "abcdbe": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertDictEqual(
            {"abcdbe": {"type": "LineString", "arcs": {0: 0, 1: 5}}}, c["objects"]
        )

    def test_cut_when_a_line_abacd_self_intersects_with_its_start_it_is_cut_into_aba_acd(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abacd": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [0, 0], [3, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertDictEqual(
            {
                "abacd": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 2, "next": {0: 2, 1: 4}},
                }
            },
            c["objects"],
        )

    def test_cut_when_a_line_abdcd_self_intersects_with_its_end_it_is_cut_into_abd_dcd(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abdcd": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [0, 0], [3, 0], [4, 0]],
                    }
                }
            )
        )

        self.assertDictEqual(
            {
                "abdcd": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 2, "next": {0: 2, 1: 4}},
                }
            },
            c["objects"],
        )

    def test_cut_when_an_old_line_abcdbe_self_intersects_and_shares_a_point_b_abcdbe_is_cut_into_ab_bcdb_be_and_fbg_is_cut_into_fb_bg(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abcdbe": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]],
                    },
                    "fbg": {"type": "LineString", "arcs": [[0, 1], [1, 0], [2, 1]]},
                }
            )
        )

        self.assertDictEqual(
            {
                "abcdbe": {
                    "type": "LineString",
                    "arcs": {0: 0, 1: 1, "next": {0: 1, 1: 4, "next": {0: 4, 1: 5}}},
                },
                "fbg": {
                    "type": "LineString",
                    "arcs": {0: 6, 1: 7, "next": {0: 7, 1: 8}},
                },
            },
            c["objects"],
        )

    def test_cut_when_a_line_abca_is_closed_there_are_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "LineString",
                        "arcs": [[0, 0], [1, 0], [0, 1], [0, 0]],
                    }
                }
            )
        )

        self.assertDictEqual(
            {"abca": {"type": "LineString", "arcs": {0: 0, 1: 3}}}, c["objects"]
        )

    def test_cut_when_a_ring_abca_is_closed_there_are_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    }
                }
            )
        )

        self.assertDictEqual(
            {"abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]}}, c["objects"]
        )

    def test_cut_exact_duplicate_rings_abca_and_abca_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "abca2": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "abca2": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_reversed_duplicate_rings_abca_and_acba_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "acba": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [0, 1], [1, 0], [0, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "acba": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_coincident_rings_abca_and_bcab_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "bcab": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [0, 1], [0, 0], [1, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "bcab": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_coincident_rings_abca_and_bacb_have_no_cuts(self):
        c = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "bacb": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [0, 0], [0, 1], [1, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "bacb": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            c["objects"],
        )

    def test_cut_coincident_rings_abcda_efae_and_ghcg_are_cut_into_abc_cda_efae_and_ghcg(
        self,
    ):
        topology = self.cut(
            self.extract(
                {
                    "abcda": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                    },
                    "efae": {
                        "type": "Polygon",
                        "arcs": [[[0, -1], [1, -1], [0, 0], [0, -1]]],
                    },
                    "ghcg": {
                        "type": "Polygon",
                        "arcs": [[[0, 2], [1, 2], [1, 1], [0, 2]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abcda": {
                    "type": "Polygon",
                    "arcs": [{0: 0, 1: 2, "next": {0: 2, 1: 4}}],
                },
                "efae": {"type": "Polygon", "arcs": [{0: 5, 1: 8}]},
                "ghcg": {"type": "Polygon", "arcs": [{0: 9, 1: 12}]},
            },
            topology["objects"],
        )

    def test_cut_coincident_rings_abca_and_dbed_have_no_cuts_but_are_rotated_to_share_b(
        self,
    ):
        topology = self.cut(
            self.extract(
                {
                    "abca": {
                        "type": "Polygon",
                        "arcs": [[[0, 0], [1, 0], [0, 1], [0, 0]]],
                    },
                    "dbed": {
                        "type": "Polygon",
                        "arcs": [[[2, 1], [1, 0], [2, 2], [2, 1]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abca": {"type": "Polygon", "arcs": [{0: 0, 1: 3}]},
                "dbed": {"type": "Polygon", "arcs": [{0: 4, 1: 7}]},
            },
            topology["objects"],
        )

        self.assertCountEqual(
            topology["coordinates"][0:4], [[1, 0], [0, 1], [0, 0], [1, 0]]
        )
        self.assertCountEqual(
            topology["coordinates"][4:8], [[1, 0], [2, 2], [2, 1], [1, 0]]
        )

    def test_cut_overlapping_rings_abcda_and_befcb_are_cut_into_bc_cdab_and_befc_cb(
        self,
    ):
        c = self.cut(
            self.extract(
                {
                    "abcda": {
                        "type": "Polygon",
                        "arcs": [
                            [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]
                        ],  # rotated to BCDAB, cut BC-CDAB
                    },
                    "befcb": {
                        "type": "Polygon",
                        "arcs": [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]],
                    },
                }
            )
        )

        self.assertDictEqual(
            {
                "abcda": {
                    "type": "Polygon",
                    "arcs": [{0: 0, 1: 1, "next": {0: 1, 1: 4}}],
                },
                "befcb": {
                    "type": "Polygon",
                    "arcs": [{0: 5, 1: 8, "next": {0: 8, 1: 9}}],
                },
            },
            c["objects"],
        )



from pytopojson import commons, feature, stitch


class ExtractArcs(object):
    def __init__(self):
        self.arcs = list()
        self.geoms_by_arc = list()
        self.geom = None

    def __call__(self, topology, object, filter, *args, **kwargs):
        self.geometry(object)

        for geoms in self.geoms_by_arc:
            if filter is None:
                self.arcs.append(geoms[0]["i"])
            else:
                if filter(geoms[0]["g"], geoms[-1]["g"]):
                    self.arcs.append(geoms[0]["i"])

        return self.arcs

    def extract_0(self, i):
        j = ~i if i < 0 else i
        self.geoms_by_arc.get(j, []).append({"i": i, "g": self.geom})

    def extract_1(self, arcs):
        for arc in arcs:
            self.extract_0(arc)

    def extract_2(self, arcs):
        for arc in arcs:
            self.extract_1(arc)

    def extract_3(self, arcs):
        for arc in arcs:
            self.extract_2(arc)

    def geometry(self, o):
        self.geom = o
        _type = o["type"]
        if _type == "GeometryCollection":
            for g in o["geometries"]:
                self.geometry(g)
        elif _type == "LineString":
            self.extract_1(o["arcs"])
        elif _type == "MultiLineString" or _type == "Polygon":
            self.extract_2(o["arcs"])
        elif _type == "MultiPolygon":
            self.extract_3(o["arcs"])


class MeshArcs(object):
    def __init__(self):
        self.stitch = stitch.Stitch()
        self.extract_arcs = ExtractArcs()

    def __call__(self, topology, obj=None, filt=None, *args, **kwargs):
        if obj is not None and filt is not None:
            arcs = self.extract_arcs(topology, obj, filt)
        else:
            n = len(topology["arcs"])
            arcs = [i for i in range(n)]
        return {"type": "MultiLineString", "arcs": self.stitch(topology, arcs)}


class Mesh(object):
    def __init__(self):
        self.object = feature.Object()
        self.mesh_arcs = MeshArcs()

    def __call__(self, topology, *args, **kwargs):
        arcs = self.mesh_arcs(topology, *args)
        return self.object(topology, arcs)



from pytopojson.hash.point import hash as hash_point
from pytopojson.hash.point import equal as equal_point
from pytopojson.hash.hash import HashMap


class Dedup(object):
    """
    Given a cut topology, combines duplicate arcs.
    """

    def __init__(self):
        pass

    def __call__(self, topology, *args, **kwargs):
        self.coordinates = topology["coordinates"]
        self.lines = topology["lines"].copy()
        self.rings = topology["rings"].copy()
        self.arc_count = len(self.lines) + len(self.rings)

        topology.pop("lines", None)
        topology.pop("rings", None)

        for l in self.lines:
            line = l
            while line.get("next", False):
                self.arc_count += 1
                line = line["next"]

        for r in self.rings:
            ring = r
            while ring.get("next", False):
                self.arc_count += 1
                ring = ring["next"]

        self.arcs_by_end = HashMap(self.arc_count * 2 * 1.4, hash_point, equal_point)
        self.arcs = list()

        for l in self.lines:
            line = l

            while line:
                self.dedup_line(line)
                line = line.get("next", False)

        for r in self.rings:
            ring = r
            # arc is no longer closed

            if ring.get("next", False):
                while ring:
                    self.dedup_line(ring)
                    ring = ring.get("next", False)
            else:
                self.dedup_ring(ring)

        topology["arcs"] = self.arcs.copy()
        return topology

    def dedup_line(self, arc):
        # Does this arc match an existing arc in order?
        start_point = self.coordinates[arc[0]]
        start_arcs = self.arcs_by_end.get(start_point)

        if start_arcs:
            for start_arc in start_arcs:
                if self.equal_line(start_arc, arc):
                    arc[0] = start_arc[0]
                    arc[1] = start_arc[1]
                    return

        # Does this arc match an existing arc in reverse order?
        end_point = self.coordinates[arc[1]]
        end_arcs = self.arcs_by_end.get(end_point)

        if end_arcs:
            for end_arc in end_arcs:
                if self.reverse_equal_line(end_arc, arc):
                    arc[0] = end_arc[1]
                    arc[1] = end_arc[0]
                    return

        if start_arcs:
            start_arcs.append(arc)
        else:
            self.arcs_by_end.set(start_point, [arc])

        if end_arcs:
            end_arcs.append(arc)
        else:
            self.arcs_by_end.set(end_point, [arc])

        self.arcs.append(arc)

    def dedup_ring(self, arc):
        # Does this arc match an existing line in order, or reverse order?
        # Rings are closed, so their start point and end point is the same.
        end_point = self.coordinates[arc[0]]
        end_arcs = self.arcs_by_end.get(end_point)

        if end_arcs:
            for end_arc in end_arcs:
                if self.equal_ring(end_arc, arc):
                    arc[0] = end_arc[0]
                    arc[1] = end_arc[1]
                    return

                if self.reverse_equal_ring(end_arc, arc):
                    arc[0] = end_arc[1]
                    arc[1] = end_arc[0]
                    return

        # Otherwise, does this arc match an existing ring in order, or reverse order?
        end_point = self.coordinates[arc[0] + self.find_minimum_offset(arc)]
        end_arcs = self.arcs_by_end.get(end_point)

        if end_arcs:
            for end_arc in end_arcs:
                if self.equal_ring(end_arc, arc):
                    arc[0] = end_arc[0]
                    arc[1] = end_arc[1]
                    return

                if self.reverse_equal_ring(end_arc, arc):
                    arc[0] = end_arc[1]
                    arc[1] = end_arc[0]
                    return

        if end_arcs:
            end_arcs.append(arc)
        else:
            self.arcs_by_end.set(end_point, [arc])

        self.arcs.append(arc)

    def equal_line(self, arc_a, arc_b):
        i_a, j_a = arc_a[0], arc_a[1]
        i_b, j_b = arc_b[0], arc_b[1]

        if i_a - j_a != i_b - j_b:
            return False

        while i_a <= j_a:
            if not equal_point(self.coordinates[i_a], self.coordinates[i_b]):
                return False
            else:
                i_a += 1
                i_b += 1

        return True

    def reverse_equal_line(self, arc_a, arc_b):
        i_a, j_a = arc_a[0], arc_a[1]
        i_b, j_b = arc_b[0], arc_b[1]

        if i_a - j_a != i_b - j_b:
            return False

        while i_a <= j_a:
            if not equal_point(self.coordinates[i_a], self.coordinates[j_b]):
                return False
            else:
                i_a += 1
                j_b -= 1

        return True

    def equal_ring(self, arc_a, arc_b):
        i_a, j_a = arc_a[0], arc_a[1]
        i_b, j_b = arc_b[0], arc_b[1]
        n = j_a - i_a

        if n != j_b - i_b:
            return False

        k_a = self.find_minimum_offset(arc_a)
        k_b = self.find_minimum_offset(arc_b)

        for i in range(n):
            if not equal_point(
                self.coordinates[i_a + (i + k_a) % n],
                self.coordinates[i_b + (i + k_b) % n],
            ):
                return False

        return True

    def reverse_equal_ring(self, arc_a, arc_b):
        i_a, j_a = arc_a[0], arc_a[1]
        i_b, j_b = arc_b[0], arc_b[1]
        n = j_a - i_a

        if n != j_b - i_b:
            return False

        k_a = self.find_minimum_offset(arc_a)
        k_b = n - self.find_minimum_offset(arc_b)

        for i in range(n):
            if not equal_point(
                self.coordinates[i_a + (i + k_a) % n],
                self.coordinates[j_b - (i + k_b) % n],
            ):
                return False

        return True

    def find_minimum_offset(self, arc):
        # Rings are rotated to a consistent, but arbitrary, start point.
        # This is necessary to detect when a ring and a rotated copy are dupes.
        start, end = arc[0], arc[1]
        mid = start
        minimum = mid
        minimum_point = self.coordinates[mid]

        mid += 1
        while mid < end:
            point = self.coordinates[mid]

            if (
                point[0] < minimum_point[0]
                or point[0] == minimum_point[0]
                and point[1] < minimum_point[1]
            ):
                minimum = mid
                minimum_point = point

            mid += 1

        return minimum - start



from pytopojson import (
    bounds,
    cut,
    dedup,
    delta,
    extract,
    geometry,
    prequantize,
)
from pytopojson.hash.hash import HashMap


class Topology(object):
    """
    Constructs the TopoJSON Topology for the specified hash of features.
    Each object in the specified hash must be a GeoJSON object,
    meaning FeatureCollection, a Feature or a geometry object.
    """

    def __init__(self):
        self.bounding_box = bounds.BoundingBox()
        self.cut = cut.Cut()
        self.dedup = dedup.Dedup()
        self.delta = delta.Delta()
        self.extract = extract.Extract()
        self.geometry = geometry.Geometry()
        self.prequantize = prequantize.Prequantize()

    def __call__(self, objects, quantization=-1):
        self.index_geometry_type = {
            "GeometryCollection": self._geometry_collection_call,
            "LineString": self._line_string_call,
            "MultiLineString": self._multi_line_string_call,
            "Polygon": self._polygon_call,
            "MultiPolygon": self._multi_polygon_call,
        }

        objects = self.geometry(objects)
        self.bbox = self.bounding_box(objects)
        self.transform = (
            quantization > 0
            and self.bbox
            and self.prequantize(objects, self.bbox, quantization)
        )
        self.topology = self.dedup(self.cut(self.extract(objects)))
        self.coordinates = self.topology["coordinates"]
        self.index_by_arc = HashMap(
            len(self.topology["arcs"]) * 1.4, self.hash_arc, self.equal_arc
        )

        objects = self.topology["objects"]
        if self.bbox is not None:
            self.topology["bbox"] = self.bbox

        self.topology["arcs"] = list(
            map(
                lambda arc, i: self._slice(arc, i),
                self.topology["arcs"],
                range(len(self.topology["arcs"])),
            )
        )

        self.topology.pop("coordinates", None)
        self.coordinates = None

        for k in objects:
            self.index_geometry(objects[k])

        if self.transform:
            self.topology["transform"] = self.transform
            self.topology["arcs"] = self.delta(self.topology["arcs"])

        return self.topology

    def _slice(self, arc, i):
        self.index_by_arc.set(arc, i)
        return self.coordinates[arc[0] : arc[1] + 1]

    def _geometry_collection_call(self, o):
        list(map(self.index_geometry, o["geometries"]))

    def _line_string_call(self, o):
        o["arcs"] = self.index_arcs(o["arcs"])

    def _multi_line_string_call(self, o):
        o["arcs"] = list(map(self.index_arcs, o["arcs"]))

    def _polygon_call(self, o):
        o["arcs"] = list(map(self.index_arcs, o["arcs"]))

    def _multi_polygon_call(self, o):
        o["arcs"] = list(map(self.index_multi_arcs, o["arcs"]))

    def index_geometry(self, geometry):
        function = self.index_geometry_type.get(geometry["type"], None)
        if geometry and function:
            function(geometry)

    def index_arcs(self, arc):
        indexes = list()

        while arc:
            index = self.index_by_arc.get(arc)
            indexes.append(index if arc[0] < arc[1] else ~index)
            arc = arc.get("next", False)

        return indexes

    def index_multi_arcs(self, arcs):
        return list(map(self.index_arcs, arcs))

    @staticmethod
    def hash_arc(arc):
        i, j = arc[0], arc[1]
        if j < i:
            t = i
            i = j
            j = t

        return i + 31 * j

    @staticmethod
    def equal_arc(arc_a, arc_b):
        i_a, j_a = arc_a[0], arc_a[1]
        i_b, j_b = arc_b[0], arc_b[1]

        if j_a < i_a:
            t = i_a
            i_a = j_a
            j_a = t

        if j_b < i_b:
            t = i_b
            i_b = j_b
            j_b = t

        return i_a == i_b and j_a == j_b



import unittest

from pytopojson import bounds


class BoundsTestCase(unittest.TestCase):
    def setUp(self):
        self.bbox = bounds.BoundingBox()

    def test_bounds_computes_bounding_box(self):
        foo = {"bar": {"type": "LineString", "arcs": [[0, 0], [1, 0], [0, 2], [0, 0]]}}

        self.assertCountEqual(self.bbox(foo), [0, 0, 1, 2])

    def test_bounds_considers_points_as_well_as_arcs(self):
        foo = {
            "bar": {
                "type": "MultiPoint",
                "coordinates": [[0, 0], [1, 0], [0, 2], [0, 0]],
            }
        }

        self.assertCountEqual(self.bbox(foo), [0, 0, 1, 2])



import json
import math
import unittest

from pytopojson import feature


class FeatureTestCase(unittest.TestCase):
    def setUp(self):
        self.feature = feature.Feature()

    @staticmethod
    def _load_json_file(filepath):
        with open(filepath, "r") as src:
            return json.load(src)

    @staticmethod
    def simple_topology(obj):
        return {
            "type": "Topology",
            "transform": {"scale": [1, 1], "translate": [0, 0]},
            "objects": {"foo": obj},
            "arcs": [
                [[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]],
                [[0, 0], [1, 0], [0, 1]],
                [[1, 1], [-1, 0], [0, -1]],
                [[1, 1]],
                [[0, 0]],
            ],
        }

    def test_feature_the_geometry_type_is_preserved(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertEqual(
            self.feature(t, t["objects"]["foo"])["geometry"]["type"], "Polygon"
        )

    def test_feature_point_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "Point", "coordinates": [0, 0]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "Point", "coordinates": [0, 0]},
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_MultiPoint_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "MultiPoint", "coordinates": [[0, 0]]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "MultiPoint", "coordinates": [[0, 0]]},
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_LineString_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "LineString", "arcs": [0]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "LineString",
                    "coordinates": [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]],
                },
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_MultiLineString_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "MultiLineString", "arcs": [[0]]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "MultiLineString",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_line_strings_have_at_least_two_coordinates(self):
        t = self.simple_topology({"type": "LineString", "arcs": [3]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "LineString", "coordinates": [[1, 1], [1, 1]]},
            },
            self.feature(t, t["objects"]["foo"]),
        )

        t = self.simple_topology({"type": "MultiLineString", "arcs": [[3], [4]]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "MultiLineString",
                    "coordinates": [[[1, 1], [1, 1]], [[0, 0], [0, 0]]],
                },
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_Polygon_is_a_valid_feature_type(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
                },
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_MultiPolygon_is_a_valid_feature_type(self):
        t = self.simple_topology({"type": "MultiPolygon", "arcs": [[[0]]]})
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "MultiPolygon",
                    "coordinates": [[[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]],
                },
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_polygons_are_closed_with_at_least_four_coordinates(self):
        topology = {
            "type": "Topology",
            "transform": {"scale": [1, 1], "translate": [0, 0]},
            "objects": {
                "foo": {"type": "Polygon", "arcs": [[0]]},
                "bar": {"type": "Polygon", "arcs": [[0, 1]]},
            },
            "arcs": [[[0, 0], [1, 1]], [[1, 1], [-1, -1]]],
        }
        self.assertListEqual(
            self.feature(topology, topology["objects"]["foo"])["geometry"][
                "coordinates"
            ],
            [[[0, 0], [1, 1], [0, 0], [0, 0]]],
        )
        self.assertListEqual(
            self.feature(topology, topology["objects"]["bar"])["geometry"][
                "coordinates"
            ],
            [[[0, 0], [1, 1], [0, 0], [0, 0]]],
        )

    def test_feature_top_level_geometry_collections_are_mapped_to_feature_collections(
        self,
    ):
        t = self.simple_topology(
            {
                "type": "GeometryCollection",
                "geometries": [{"type": "MultiPolygon", "arcs": [[[0]]]}],
            }
        )
        self.assertDictEqual(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "properties": {},
                        "geometry": {
                            "type": "MultiPolygon",
                            "coordinates": [[[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]],
                        },
                    }
                ],
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_geometry_collections_can_be_nested(self):
        t = self.simple_topology(
            {
                "type": "GeometryCollection",
                "geometries": [
                    {
                        "type": "GeometryCollection",
                        "geometries": [{"type": "Point", "coordinates": [0, 0]}],
                    }
                ],
            }
        )
        self.assertDictEqual(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "properties": {},
                        "geometry": {
                            "type": "GeometryCollection",
                            "geometries": [{"type": "Point", "coordinates": [0, 0]}],
                        },
                    }
                ],
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_top_level_geometry_collections_do_not_have_ids_but_second_level_geometry_collections_can(
        self,
    ):
        t = self.simple_topology(
            {
                "type": "GeometryCollection",
                "id": "collection",
                "geometries": [
                    {
                        "type": "GeometryCollection",
                        "id": "feature",
                        "geometries": [
                            {"type": "Point", "id": "geometry", "coordinates": [0, 0]}
                        ],
                    }
                ],
            }
        )
        self.assertDictEqual(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "id": "feature",
                        "properties": {},
                        "geometry": {
                            "type": "GeometryCollection",
                            "geometries": [{"type": "Point", "coordinates": [0, 0]}],
                        },
                    }
                ],
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_top_level_geometry_collections_do_not_have_properties_but_second_level_geometry_collections_can(
        self,
    ):
        t = self.simple_topology(
            {
                "type": "GeometryCollection",
                "properties": {"collection": True},
                "geometries": [
                    {
                        "type": "GeometryCollection",
                        "properties": {"feature": True},
                        "geometries": [
                            {
                                "type": "Point",
                                "properties": {"geometry": True},
                                "coordinates": [0, 0],
                            }
                        ],
                    }
                ],
            }
        )
        self.assertDictEqual(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "properties": {"feature": True},
                        "geometry": {
                            "type": "GeometryCollection",
                            "geometries": [{"type": "Point", "coordinates": [0, 0]}],
                        },
                    }
                ],
            },
            self.feature(t, t["objects"]["foo"]),
        )

    def test_feature_the_object_id_is_promoted_to_feature_id(self):
        t = self.simple_topology({"id": "foo", "type": "Polygon", "arcs": [[0]]})
        self.assertEqual("foo", self.feature(t, t["objects"]["foo"])["id"])

    def test_feature_any_object_properties_are_promoted_to_feature_properties(self):
        t = self.simple_topology(
            {
                "type": "Polygon",
                "properties": {"color": "orange", "size": 42},
                "arcs": [[0]],
            }
        )
        self.assertDictEqual(
            {"color": "orange", "size": 42},
            self.feature(t, t["objects"]["foo"])["properties"],
        )

    def test_feature_the_object_id_is_optional(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertEqual(self.feature(t, t["objects"]["foo"]).get("id", None), None)

    def test_feature_object_properties_are_created_if_missing(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertDictEqual(self.feature(t, t["objects"]["foo"])["properties"], {})

    def test_feature_arcs_are_converted_to_coordinates(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertListEqual(
            self.feature(t, t["objects"]["foo"])["geometry"]["coordinates"],
            [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]],
        )

    def test_feature_negative_arc_indexes_indicate_reversed_coordinates(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[~0]]})
        self.assertListEqual(
            self.feature(t, t["objects"]["foo"])["geometry"]["coordinates"],
            [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]],
        )

    def test_feature_when_multiple_arc_indexes_are_specified_coordinates_are_stitched_together(
        self,
    ):
        t = self.simple_topology({"type": "LineString", "arcs": [1, 2]})
        self.assertListEqual(
            self.feature(t, t["objects"]["foo"])["geometry"]["coordinates"],
            [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]],
        )

        t = self.simple_topology({"type": "Polygon", "arcs": [[~2, ~1]]})
        self.assertListEqual(
            self.feature(t, t["objects"]["foo"])["geometry"]["coordinates"],
            [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]],
        )

    def test_feature_unknown_geometry_types_are_converted_to_null_geometries(self):
        topology = {
            "type": "Topology",
            "transform": {"scale": [1, 1], "translate": [0, 0]},
            "objects": {
                "foo": {"id": "foo"},
                "bar": {"type": "Invalid", "properties": {"bar": 2}},
                "baz": {
                    "type": "GeometryCollection",
                    "geometries": [{"type": "Unknown", "id": "unknown"}],
                },
            },
            "arcs": [],
        }
        self.assertDictEqual(
            {"type": "Feature", "id": "foo", "properties": {}, "geometry": None},
            self.feature(topology, topology["objects"]["foo"]),
        )
        self.assertDictEqual(
            {"type": "Feature", "properties": {"bar": 2}, "geometry": None},
            self.feature(topology, topology["objects"]["bar"]),
        )
        self.assertDictEqual(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "id": "unknown",
                        "properties": {},
                        "geometry": None,
                    }
                ],
            },
            self.feature(topology, topology["objects"]["baz"]),
        )

    def test_feature_preserves_additional_dimensions_in_Point_geometries(self):
        t = {
            "type": "Topology",
            "objects": {"point": {"type": "Point", "coordinates": [1, 2, "foo"]}},
            "arcs": [],
        }
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "Point", "coordinates": [1, 2, "foo"]},
            },
            self.feature(t, t["objects"]["point"]),
        )

    def test_feature_preserves_additional_dimensions_in_MultiPoint_geometries(self):
        t = {
            "type": "Topology",
            "objects": {
                "points": {"type": "MultiPoint", "coordinates": [[1, 2, "foo"]]}
            },
            "arcs": [],
        }
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {"type": "MultiPoint", "coordinates": [[1, 2, "foo"]]},
            },
            self.feature(t, t["objects"]["points"]),
        )

    def test_feature_preserves_additional_dimensions_in_LineString_geometries(self):
        t = {
            "type": "Topology",
            "objects": {"line": {"type": "LineString", "arcs": [0]}},
            "arcs": [[[1, 2, "foo"], [3, 4, "bar"]]],
        }
        self.assertDictEqual(
            {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "LineString",
                    "coordinates": [[1, 2, "foo"], [3, 4, "bar"]],
                },
            },
            self.feature(t, t["objects"]["line"]),
        )



import json
import unittest

from pytopojson import quantize


class QuantizeTestCase(unittest.TestCase):
    def setUp(self):
        self.quantize = quantize.Quantize()

    @staticmethod
    def _load_json_file(filepath):
        with open(filepath, "r") as src:
            return json.load(src)

    def test_quantize_the_input_topology(self):
        polygon = self._load_json_file("tests/client/topojson/polygon.json")

        quantized = self._load_json_file("tests/client/topojson/polygon-q1e4.json")
        self.assertDictEqual(self.quantize(polygon, 1e4), quantized)

        quantized = self._load_json_file("tests/client/topojson/polygon-q1e5.json")
        self.assertDictEqual(self.quantize(polygon, 1e5), quantized)

    def test_quantize_ensures_that_each_arc_has_at_least_two_points(self):
        polygon = self._load_json_file("tests/client/topojson/empty.json")
        quantized = self._load_json_file("tests/client/topojson/empty-q1e4.json")
        self.assertDictEqual(self.quantize(polygon, 1e4), quantized)

    def test_quantize_preserves_the_id_bbox_and_properties_of_input_objects(self):
        polygon = self._load_json_file("tests/client/topojson/properties.json")
        quantized = self._load_json_file("tests/client/topojson/properties-q1e4.json")
        self.assertDictEqual(self.quantize(polygon, 1e4), quantized)

    def test_quantize_throws_an_error_if_n_is_not_at_least_two(self):
        topology = self._load_json_file("tests/client/topojson/polygon.json")
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology, 0)
        self.assertEqual("n must be ≥2.", str(exc.exception))
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology, 1.5)
        self.assertEqual("n must be ≥2.", str(exc.exception))
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology)
        self.assertEqual("n must be ≥2.", str(exc.exception))
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology, None)
        self.assertEqual("n must be ≥2.", str(exc.exception))
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology, -2)
        self.assertEqual("n must be ≥2.", str(exc.exception))

    def test_quantize_throws_an_error_if_the_topology_is_already_quantized(self):
        topology = self._load_json_file("tests/client/topojson/polygon-q1e4.json")
        with self.assertRaises(ValueError) as exc:
            self.quantize(topology, 1e4)
        self.assertEqual("Already quantized.", str(exc.exception))

    def test_quantize_returns_a_new_topology_with_a_bounding_box(self):
        before = self._load_json_file("tests/client/topojson/polygon.json")
        before["bbox"] = None
        after = self.quantize(before, 1e4)

        quantized = self._load_json_file("tests/client/topojson/polygon-q1e4.json")
        self.assertDictEqual(after, quantized)
        self.assertListEqual(after["bbox"], [0, 0, 10, 10])
        self.assertIsNone(before["bbox"])
